@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of urlwatch (https://thp.io/2008/urlwatch/).
-# Copyright (c) 2008-2019 Thomas Perl <m@thp.io>
+# Copyright (c) 2008-2021 Thomas Perl <m@thp.io>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -32,6 +32,7 @@ import os
 import stat
 import copy
 import platform
+import collections
 from abc import ABCMeta, abstractmethod
 
 import shutil
@@ -39,8 +40,24 @@ import yaml
 import minidb
 import logging
 
+try:
+    import msgpack
+except ImportError:
+    msgpack = None
+
+try:
+    import redis
+except ImportError:
+    redis = None
+
+try:
+    import pwd
+except ImportError:
+    pwd = None
+
 from .util import atomic_rename, edit_file
 from .jobs import JobBase, UrlJob, ShellJob
+from .filters import FilterBase
 
 logger = logging.getLogger(__name__)
 
@@ -59,6 +76,12 @@ DEFAULT_CONFIG = {
             'minimal': False,
         },
 
+        'markdown': {
+            'details': True,
+            'footer': True,
+            'minimal': False,
+        },
+
         'html': {
             'diff': 'unified',  # "unified" or "table"
         },
@@ -78,9 +101,10 @@ DEFAULT_CONFIG = {
             'method': 'smtp',
             'smtp': {
                 'host': 'localhost',
+                'user': '',
                 'port': 25,
                 'starttls': True,
-                'keyring': True,
+                'auth': True,
             },
             'sendmail': {
                 'path': 'sendmail',
@@ -89,9 +113,10 @@ DEFAULT_CONFIG = {
         'pushover': {
             'enabled': False,
             'app': '',
-            'device': '',
+            'device': None,
             'sound': 'spacealarm',
             'user': '',
+            'priority': 'normal',
         },
         'pushbullet': {
             'enabled': False,
@@ -101,10 +126,31 @@ DEFAULT_CONFIG = {
             'enabled': False,
             'bot_token': '',
             'chat_id': '',
+            'monospace': False,
+            'silent': False,
         },
         'slack': {
             'enabled': False,
             'webhook_url': '',
+            'max_message_length': 40000,
+        },
+        'mattermost': {
+            'enabled': False,
+            'webhook_url': '',
+            'max_message_length': 40000,
+        },
+        'discord': {
+            'enabled': False,
+            'embed': False,
+            'subject': '{count} changes: {jobs}',
+            'webhook_url': '',
+            'max_message_length': 2000,
+        },
+        'matrix': {
+            'enabled': False,
+            'homeserver': '',
+            'access_token': '',
+            'room_id': '',
         },
         'mailgun': {
             'enabled': False,
@@ -116,6 +162,23 @@ DEFAULT_CONFIG = {
             'to': '',
             'subject': '{count} changes: {jobs}'
         },
+        'ifttt': {
+            'enabled': False,
+            'key': '',
+            'event': '',
+        },
+        'xmpp': {
+            'enabled': False,
+            'sender': '',
+            'recipient': '',
+        },
+        'prowl': {
+            'enabled': False,
+            'api_key': '',
+            'priority': 0,
+            'application': '',
+            'subject': '{count} changes: {jobs}'
+        },
     },
 
     'job_defaults': {
@@ -147,7 +210,6 @@ def get_current_user():
         # If there is no controlling terminal, because urlwatch is launched by
         # cron, or by a systemd.service for example, os.getlogin() fails with:
         # OSError: [Errno 25] Inappropriate ioctl for device
-        import pwd
         return pwd.getpwuid(os.getuid()).pw_name
 
 
@@ -184,6 +246,7 @@ class BaseTextualFileStorage(BaseFileStorage, metaclass=ABCMeta):
         if os.path.exists(self.filename):
             shutil.copy(self.filename, file_edit)
         elif example_file is not None and os.path.exists(example_file):
+            os.makedirs(os.path.dirname(file_edit) or '.', exist_ok=True)
             shutil.copy(example_file, file_edit)
 
         while True:
@@ -214,6 +277,7 @@ class BaseTextualFileStorage(BaseFileStorage, metaclass=ABCMeta):
 
     @classmethod
     def write_default_config(cls, filename):
+        os.makedirs(os.path.dirname(filename) or '.', exist_ok=True)
         config_storage = cls(None)
         config_storage.filename = filename
         config_storage.save()
@@ -249,12 +313,25 @@ class UrlsBaseFileStorage(BaseTextualFileStorage, metaclass=ABCMeta):
     def load_secure(self):
         jobs = self.load()
 
+        def is_shell_job(job):
+            if isinstance(job, ShellJob):
+                return True
+
+            for filter_kind, subfilter in FilterBase.normalize_filter_list(job.filter):
+                if filter_kind == 'shellpipe':
+                    return True
+
+                if job.diff_tool is not None:
+                    return True
+
+            return False
+
         # Security checks for shell jobs - only execute if the current UID
         # is the same as the file/directory owner and only owner can write
         shelljob_errors = self.shelljob_security_checks()
-        if shelljob_errors and any(isinstance(job, ShellJob) for job in jobs):
+        if shelljob_errors and any(is_shell_job(job) for job in jobs):
             print(('Removing shell jobs, because %s' % (' and '.join(shelljob_errors),)))
-            jobs = [job for job in jobs if not isinstance(job, ShellJob)]
+            jobs = [job for job in jobs if not is_shell_job(job)]
 
         return jobs
 
@@ -288,7 +365,7 @@ class BaseYamlFileStorage(BaseTextualFileStorage, metaclass=ABCMeta):
         filename = args[0]
         if filename is not None and os.path.exists(filename):
             with open(filename) as fp:
-                return yaml.load(fp)
+                return yaml.load(fp, Loader=yaml.SafeLoader)
 
 
 class YamlConfigStorage(BaseYamlFileStorage):
@@ -301,13 +378,31 @@ class YamlConfigStorage(BaseYamlFileStorage):
 
 
 class UrlsYaml(BaseYamlFileStorage, UrlsBaseFileStorage):
+    @classmethod
+    def _parse(cls, fp):
+        jobs = [JobBase.unserialize(job) for job in yaml.load_all(fp, Loader=yaml.SafeLoader)
+                if job is not None]
+        jobs_by_guid = collections.defaultdict(list)
+        for job in jobs:
+            jobs_by_guid[job.get_guid()].append(job)
+
+        conflicting_jobs = []
+        for guid, guid_jobs in jobs_by_guid.items():
+            if len(guid_jobs) != 1:
+                conflicting_jobs.append(guid_jobs[0].get_location())
+
+        if conflicting_jobs:
+            raise ValueError('\n   '.join(['Each job must have a unique URL, append #1, #2, ... to make them unique:']
+                                          + conflicting_jobs))
+
+        return jobs
 
     @classmethod
     def parse(cls, *args):
         filename = args[0]
         if filename is not None and os.path.exists(filename):
             with open(filename) as fp:
-                return [JobBase.unserialize(job) for job in yaml.load_all(fp) if job is not None]
+                return cls._parse(fp)
 
     def save(self, *args):
         jobs = args[0]
@@ -318,7 +413,7 @@ class UrlsYaml(BaseYamlFileStorage, UrlsBaseFileStorage):
 
     def load(self, *args):
         with open(self.filename) as fp:
-            return [JobBase.unserialize(job) for job in yaml.load_all(fp) if job is not None]
+            return self._parse(fp)
 
 
 class UrlsTxt(BaseTxtFileStorage, UrlsBaseFileStorage):
@@ -450,13 +545,31 @@ class CacheMiniDBStorage(CacheStorage):
         return (guid for guid, in CacheEntry.query(self.db, minidb.Function('distinct', CacheEntry.c.guid)))
 
     def load(self, job, guid):
-        for data, timestamp, tries, etag in CacheEntry.query(self.db, CacheEntry.c.data // CacheEntry.c.timestamp // CacheEntry.c.tries // CacheEntry.c.etag,
-                                                             order_by=minidb.columns(CacheEntry.c.timestamp.desc, CacheEntry.c.tries.desc),
+        for data, timestamp, tries, etag in CacheEntry.query(self.db,
+                                                             CacheEntry.c.data // CacheEntry.c.timestamp
+                                                             // CacheEntry.c.tries // CacheEntry.c.etag,
+                                                             order_by=minidb.columns(CacheEntry.c.timestamp.desc,
+                                                                                     CacheEntry.c.tries.desc),
                                                              where=CacheEntry.c.guid == guid, limit=1):
             return data, timestamp, tries, etag
 
         return None, None, 0, None
 
+    def get_history_data(self, guid, count=1):
+        history = {}
+        if count < 1:
+            return history
+        for data, timestamp in CacheEntry.query(self.db, CacheEntry.c.data // CacheEntry.c.timestamp,
+                                                order_by=minidb.columns(CacheEntry.c.timestamp.desc,
+                                                                        CacheEntry.c.tries.desc),
+                                                where=(CacheEntry.c.guid == guid)
+                                                & ((CacheEntry.c.tries == 0) | (CacheEntry.c.tries == None))):  # noqa:E711
+            if data not in history:
+                history[data] = timestamp
+                if len(history) >= count:
+                    break
+        return history
+
     def save(self, job, guid, data, timestamp, tries, etag=None):
         self.db.save(CacheEntry(guid=guid, timestamp=timestamp, data=data, tries=tries, etag=etag))
         self.db.commit()
@@ -475,3 +588,72 @@ class CacheMiniDBStorage(CacheStorage):
             return result
 
         return 0
+
+
+class CacheRedisStorage(CacheStorage):
+    def __init__(self, filename):
+        super().__init__(filename)
+
+        if redis is None or msgpack is None:
+            raise ImportError('redis + msgpack are missing')
+
+        self.db = redis.from_url(filename)
+
+    def _make_key(self, guid):
+        return 'guid:' + guid
+
+    def close(self):
+        self.db.connection_pool.disconnect()
+        self.db = None
+
+    def get_guids(self):
+        guids = []
+        for guid in self.db.keys(b'guid:*'):
+            guids.append(str(guid[len('guid:'):]))
+        return guids
+
+    def load(self, job, guid):
+        key = self._make_key(guid)
+        data = self.db.lindex(key, 0)
+
+        if data:
+            r = msgpack.unpackb(data)
+            return r['data'], r['timestamp'], r['tries'], r['etag']
+
+        return None, None, 0, None
+
+    def get_history_data(self, guid, count=1):
+        history = {}
+        if count < 1:
+            return history
+
+        key = self._make_key(guid)
+        for i in range(0, self.db.llen(key)):
+            r = self.db.lindex(key, i)
+            c = msgpack.unpackb(r)
+            if (c['tries'] == 0 or c['tries'] is None):
+                if c['data'] not in history:
+                    history[c['data']] = c['timestamp']
+                    if len(history) >= count:
+                        break
+        return history
+
+    def save(self, job, guid, data, timestamp, tries, etag=None):
+        r = {
+            'data': data,
+            'timestamp': timestamp,
+            'tries': tries,
+            'etag': etag,
+        }
+        self.db.lpush(self._make_key(guid), msgpack.packb(r, use_bin_type=True))
+
+    def delete(self, guid):
+        self.db.delete(self._make_key(guid))
+
+    def clean(self, guid):
+        key = self._make_key(guid)
+        i = self.db.llen(key)
+        if self.db.ltrim(key, 0, 0):
+            return i - self.db.llen(key)
+
+        return 0
